from dataclasses import dataclass
import logging
from time import time
from argus.backend.plugins.sct.testrun import SCTTestRun
from argus.backend.plugins.sct.udt import (
    CloudInstanceDetails,
    CloudResource,
    EventsBySeverity,
    NemesisRunInfo,
    NodeDescription,
    PackageVersion,
)
from argus.backend.util.enums import NemesisStatus, ResourceState

LOGGER = logging.getLogger(__name__)


class SCTServiceException(Exception):
    pass


@dataclass(init=True, repr=True)
class NemesisSubmissionRequest:
    name: str
    class_name: str
    start_time: int
    node_name: str
    node_ip: str
    node_shards: int


@dataclass(init=True, repr=True)
class NemesisFinalizationRequest:
    name: str
    start_time: int
    status: str
    message: str


@dataclass(init=True, repr=True)
class EventSubmissionRequest:
    severity: str
    total_events: int
    messages: list[str]


class SCTService:

    @staticmethod
    def submit_packages(run_id: str, packages: list[dict]) -> str:
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            for package_dict in packages:
                package = PackageVersion(**package_dict)
                run.packages.append(package)
            run.save()
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "added"

    @staticmethod
    def submit_screenshots(run_id: str, screenshot_links: list[str]) -> str:
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            for link in screenshot_links:
                run.add_screenshot(link)
            run.save()
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "submitted"

    @staticmethod
    def create_resource(run_id: str, resource_details: dict) -> str:
        instance_details = CloudInstanceDetails(**resource_details.pop("instance_details"))
        resource = CloudResource(**resource_details, instance_info=instance_details)
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            run.get_resources().append(resource)
            run.save()
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "created"

    @staticmethod
    def update_resource_shards(run_id: str, resource_name: str, new_shards: int) -> str:
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            resource = next(res for res in run.get_resources() if res.name == resource_name)
            resource.get_instance_info().shards_amount = new_shards
            run.save()
        except StopIteration as exception:
            LOGGER.error("Resource %s not found in run %s", resource_name, run_id)
            raise SCTServiceException("Resource not found", resource_name) from exception
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "updated"

    @staticmethod
    def terminate_resource(run_id: str, resource_name: str, reason: str) -> str:
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            resource = next(res for res in run.get_resources() if res.name == resource_name)
            resource.get_instance_info().termination_reason = reason
            resource.get_instance_info().termination_time = int(time())
            resource.state = ResourceState.TERMINATED.value
            run.save()
        except StopIteration as exception:
            LOGGER.error("Resource %s not found in run %s", resource_name, run_id)
            raise SCTServiceException("Resource not found", resource_name) from exception
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "terminated"

    @staticmethod
    def submit_nemesis(run_id: str, nemesis_details: dict) -> str:
        nem_req = NemesisSubmissionRequest(**nemesis_details)
        node_desc = NodeDescription(name=nem_req.node_name, ip=nem_req.node_ip, shards=nem_req.node_shards)
        nemesis_info = NemesisRunInfo(
            class_name=nem_req.class_name,
            name=nem_req.name,
            start_time=int(nem_req.start_time),
            end_time=0,
            duration=0,
            stack_trace="",
            status=NemesisStatus.RUNNING.value,
            target_node=node_desc,
        )
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            run.add_nemesis(nemesis_info)
            run.save()
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "created"

    @staticmethod
    def finalize_nemesis(run_id: str, nemesis_details: dict) -> str:
        nem_req = NemesisFinalizationRequest(**nemesis_details)
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            nemesis = next(nem for nem in run.get_nemeses() if nem.name ==
                           nem_req.name and nem.start_time == nem_req.start_time)
            nemesis.status = NemesisStatus(nem_req.status).value
            nemesis.stack_trace = nem_req.message
            nemesis.end_time = int(time())
            run.save()
        except StopIteration as exception:
            LOGGER.error("Nemesis %s (%s) not found for run %s", nem_req.name, nem_req.start_time, run_id)
            raise SCTServiceException("Nemesis not found", (nem_req.name, nem_req.start_time)) from exception
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "updated"

    @staticmethod
    def submit_events(run_id: str, events: list[dict]) -> str:
        wrapped_events = [EventSubmissionRequest(**ev) for ev in events]
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            for event in wrapped_events:
                wrapper = EventsBySeverity(severity=event.severity,
                                           event_amount=event.total_events, last_events=event.messages)
                run.get_events().append(wrapper)
            run.save()
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "added"

    @staticmethod
    def get_scylla_version_kernels_report(release_name: str):
        all_release_runs = SCTTestRun.get_version_data_for_release(release_name=release_name)
        kernels_by_version = {}
        kernel_metadata = {}
        for run in all_release_runs:
            packages = run["packages"]
            if not packages:
                continue
            scylla_pkgs = {p["name"]: p for p in packages if "scylla-server" in p["name"]}
            scylla_pkg = scylla_pkgs["scylla-server-upgraded"] if scylla_pkgs.get(
                "scylla-server-upgraded") else scylla_pkgs.get("scylla-server")
            version = f"{scylla_pkg['version']}-{scylla_pkg['date']}.{scylla_pkg['revision_id']}" if scylla_pkgs else "unknown"
            kernel_packages = [p for p in packages if "kernel" in p["name"]]
            kernel_package = kernel_packages[0] if len(kernel_packages) > 0 else None
            if not kernel_package:
                continue
            version_list = set(kernels_by_version.get(version, []))
            version_list.add(kernel_package["version"])
            kernels_by_version[version] = list(version_list)
            metadata = kernel_metadata.get(
                kernel_package.version,
                {
                    "passed": 0,
                    "failed": 0,
                    "aborted": 0,
                }
            )
            if run["status"] in ["passed", "failed", "aborted"]:
                metadata[run["status"]] += 1
            kernel_metadata[kernel_package["version"]] = metadata

        return {
            "versions": kernels_by_version,
            "metadata": kernel_metadata
        }



from datetime import datetime
import functools
import hashlib
import os
import base64
from uuid import UUID
from time import time
from hashlib import sha384

from flask import current_app, flash, g, redirect, request, session, url_for
import requests
from werkzeug.security import generate_password_hash, check_password_hash

from argus.backend.db import ScyllaCluster
from argus.backend.error_handlers import APIException
from argus.backend.models.web import User, UserOauthToken, UserRoles, WebFileStorage
from argus.backend.util.common import FlaskView


class GithubOrganizationMissingError(Exception):
    pass


class UserService:
    # pylint: disable=no-self-use
    def __init__(self) -> None:
        self.cluster = ScyllaCluster.get()
        self.session = self.cluster.session

    @staticmethod
    def check_roles(roles: list[UserRoles] | UserRoles, user: User) -> bool:
        if isinstance(roles, str):
            return roles in user.roles
        elif isinstance(roles, list):
            for role in roles:
                if role in user.roles:
                    return True
        return False

    def github_callback(self, req_code: str) -> dict | None:
        # pylint: disable=too-many-locals
        oauth_response = requests.post(
            "https://github.com/login/oauth/access_token",
            headers={
                "Accept": "application/json",
            },
            params={
                "code": req_code,
                "client_id": current_app.config.get("GITHUB_CLIENT_ID"),
                "client_secret": current_app.config.get("GITHUB_CLIENT_SECRET"),
            }
        )

        oauth_data = oauth_response.json()

        user_info = requests.get(
            "https://api.github.com/user",
            headers={
                "Accept": "application/json",
                "Authorization": f"token {oauth_data.get('access_token')}"
            }
        ).json()
        email_info = requests.get(
            "https://api.github.com/user/emails",
            headers={
                "Accept": "application/json",
                "Authorization": f"token {oauth_data.get('access_token')}"
            }
        ).json()

        organizations = requests.get(
            "https://api.github.com/user/orgs",
            headers={
                "Accept": "application/json",
                "Authorization": f"token {oauth_data.get('access_token')}"
            }
        ).json()

        temp_password = None
        required_organizations = current_app.config.get("GITHUB_REQUIRED_ORGANIZATIONS")
        if required_organizations:
            logins = set([org["login"] for org in organizations])  # pylint: disable=consider-using-set-comprehension
            required_organizations = set(required_organizations)
            if len(logins.intersection(required_organizations)) == 0:
                raise GithubOrganizationMissingError(
                    "Not a member of a required organization or missing organization scope")

        try:
            user = User.get(username=user_info.get("login"))
        except User.DoesNotExist:
            user = User()
            user.username = user_info.get("login")
            user.email = email_info[-1].get("email")
            user.full_name = user_info.get("name", user_info.get("login"))
            user.registration_date = datetime.utcnow()
            user.roles = ["ROLE_USER"]
            temp_password = base64.encodebytes(
                os.urandom(48)).decode("ascii").strip()
            user.password = generate_password_hash(temp_password)

            avatar_url: str = user_info.get("avatar_url")
            avatar = requests.get(avatar_url).content
            avatar_name = avatar_url.split("/")[-1]
            filename, filepath = self.save_profile_picture_to_disk(avatar_name, avatar, user.username)

            web_file = WebFileStorage()
            web_file.filename = filename
            web_file.filepath = filepath
            web_file.save()
            user.picture_id = web_file.id
            user.save()

        try:
            tokens = list(UserOauthToken.filter(user_id=user.id).all())
            github_token = [
                token for token in tokens if token["kind"] == "github"][0]
            github_token.token = oauth_data.get('access_token')
            github_token.save()
        except (UserOauthToken.DoesNotExist, IndexError):
            github_token = UserOauthToken()
            github_token.kind = "github"
            github_token.user_id = user.id
            github_token.token = oauth_data.get('access_token')
            github_token.save()

        session.clear()
        session["user_id"] = str(user.id)
        if temp_password:
            return {
                "password": temp_password,
                "first_login": True
            }
        return None

    def get_users(self) -> dict:
        users = User.all()
        return {str(user.id): user.to_json() for user in users}

    def generate_token(self, user: User):
        token_digest = f"{user.username}-{int(time())}-{base64.encodebytes(os.urandom(128)).decode(encoding='utf-8')}"
        new_token = base64.encodebytes(sha384(token_digest.encode(encoding="utf-8")
                                              ).digest()).decode(encoding="utf-8").strip()
        user.api_token = new_token
        user.save()
        return new_token

    def update_email(self, user: User, new_email: str):
        user.email = new_email
        user.save()

    def update_password(self, user: User, old_password: str, new_password: str):
        if check_password_hash(user.password, old_password):
            raise Exception("Incorrect old password")

        user.password = generate_password_hash(new_password)
        user.save()

    def update_name(self, user: User, new_name: str):
        user.full_name = new_name
        user.save()

    def save_profile_picture_to_disk(self, original_filename: str, filedata: bytes, suffix: str):
        filename_fragment = hashlib.sha256(os.urandom(64)).hexdigest()[:10]
        filename = f"profile_{suffix}_{filename_fragment}"
        filepath = f"storage/profile_pictures/{filename}"
        with open(filepath, "wb") as file:
            file.write(filedata)

        return original_filename, filepath

    def update_profile_picture(self, filename: str, filepath: str):
        web_file = WebFileStorage()
        web_file.filename = filename
        web_file.filepath = filepath
        web_file.save()

        try:
            if old_picture_id := g.user.picture_id:
                old_file = WebFileStorage.get(id=old_picture_id)
                os.unlink(old_file.filepath)
                old_file.delete()
        except Exception as exc:  # pylint: disable=broad-except
            print(exc)

        g.user.picture_id = web_file.id
        g.user.save()


def login_required(view: FlaskView):
    @functools.wraps(view)
    def wrapped_view(*args, **kwargs):
        if g.user is None and not getattr(view, "api_view", False):
            flash(message='Unauthorized, please login', category='error')
            return redirect(url_for('auth.login'))
        elif g.user is None and getattr(view, "api_view", True):
            return {
                "status": "error",
                "message": "Authorization required"
            }, 403

        return view(*args, **kwargs)

    return wrapped_view


def api_login_required(view: FlaskView):
    view.api_view = True
    return login_required(view)


def check_roles(needed_roles: list[str] | str = None):
    def inner(view: FlaskView):
        @functools.wraps(view)
        def wrapped_view(*args, **kwargs):
            if not UserService.check_roles(needed_roles, g.user):
                flash(message='Not authorized to access this area', category='error')
                return redirect(url_for('main.home'))

            return view(*args, **kwargs)

        return wrapped_view
    return inner


def load_logged_in_user():
    user_id = session.get('user_id')
    auth_header = request.headers.get("Authorization")

    if user_id:
        try:
            g.user = User.get(id=UUID(user_id))  # pylint: disable=assigning-non-slot
            return
        except User.DoesNotExist:
            session.clear()

    if auth_header:
        try:
            auth_schema, *auth_data = auth_header.split()
            if auth_schema == "token":
                token = auth_data[0]
                g.user = User.get(api_token=token)
                return
        except IndexError as exception:
            raise APIException("Malformed authorization header") from exception
        except User.DoesNotExist as exception:
            raise APIException("User not found for supplied token") from exception
    g.user = None  # pylint: disable=assigning-non-slot



import logging
from flask import (
    Blueprint,
    g,
    request,
)
from argus.backend.error_handlers import handle_api_exception
from argus.backend.service.notification_manager import NotificationManagerService
from argus.backend.service.user import api_login_required
from argus.backend.util.common import get_payload

bp = Blueprint('notifications', __name__, url_prefix='/notifications')
LOGGER = logging.getLogger(__name__)
bp.register_error_handler(Exception, handle_api_exception)


@bp.route("/get")
@api_login_required
def get_notification():
    notification_id = request.args.get("id")
    if not notification_id:
        raise Exception("No notification id provided")
    service = NotificationManagerService()
    notification = service.get_notificaton(
        receiver=g.user.id, notification_id=notification_id)
    return {
        "status": "ok",
        "response": notification.to_dict()
    }


@bp.route("/get_unread")
@api_login_required
def get_unread_count():
    service = NotificationManagerService()
    unread_count = service.get_unread_count(receiver=g.user.id)
    return {
        "status": "ok",
        "response": unread_count
    }


@bp.route("/summary")
@api_login_required
def get_summary():
    after = request.args.get("afterId")
    limit = request.args.get("limit")
    limit = int(limit) if limit else 20
    service = NotificationManagerService()
    notifications = service.get_notifications(
        receiver=g.user.id,
        limit=limit,
        after=after
    )
    return {
        "status": "ok",
        "response": [n.to_dict_short_summary() for n in notifications]
    }


@bp.route("/read", methods=["POST"])
@api_login_required
def read_notification():
    payload = get_payload(request)
    service = NotificationManagerService()
    status = service.read_notification(
        receiver=g.user.id, notification_id=payload["id"])

    return {
        "status": "ok",
        "response": status
    }



import logging
from datetime import datetime
from dataclasses import dataclass
from uuid import UUID

from cassandra.cqlengine import columns
from argus.backend.db import ScyllaCluster
from argus.backend.models.web import ArgusRelease
from argus.backend.plugins.core import PluginModelBase
from argus.backend.plugins.sct.resource_setup import ResourceSetup
from argus.backend.plugins.sct.udt import (
    CloudInstanceDetails,
    CloudResource,
    CloudSetupDetails,
    EventsBySeverity,
    NemesisRunInfo,
    PackageVersion,
)

LOGGER = logging.getLogger(__name__)
SCT_REGION_PROPERTY_MAP = {
    "aws": "region_name",
    "aws-siren": "region_name",
    "k8s-eks": "region_name",
    "gce": "gce_datacenter",
    "gce-siren": "gce_datacenter",
    "k8s-gke": "gce_datacenter",
    "azure": "azure_region_name",
    "default": "region_name",
}


@dataclass(init=True, repr=True, frozen=True)
class SCTTestRunSubmissionRequest():
    schema_version: str
    run_id: str
    job_name: str
    job_url: str
    started_by: str
    commit_id: str
    runner_public_ip: str
    runner_private_ip: str
    sct_config: dict


class SCTTestRun(PluginModelBase):
    __table_name__ = "test_runs_v8"
    _plugin_name = "scylla-cluster-tests"

    # Test Details
    scm_revision_id = columns.Text()
    started_by = columns.Text()
    config_files = columns.List(value_type=columns.Text())
    packages = columns.List(value_type=columns.UserDefinedType(user_type=PackageVersion))
    scylla_version = columns.Text()
    yaml_test_duration = columns.Integer()

    # Test Preset Resources
    sct_runner_host = columns.UserDefinedType(user_type=CloudInstanceDetails)
    region_name = columns.List(value_type=columns.Text())
    cloud_setup = columns.UserDefinedType(user_type=CloudSetupDetails)

    # Test Runtime Resources
    allocated_resources = columns.List(value_type=columns.UserDefinedType(user_type=CloudResource))

    # Test Results
    events = columns.List(value_type=columns.UserDefinedType(user_type=EventsBySeverity))
    nemesis_data = columns.List(value_type=columns.UserDefinedType(user_type=NemesisRunInfo))
    screenshots = columns.List(value_type=columns.Text())

    @classmethod
    def _stats_query(cls) -> str:
        return ("SELECT id, test_id, group_id, release_id, status, start_time, build_job_url, build_id, "
                f"assignee, end_time, investigation_status, heartbeat, scylla_version FROM {cls.table_name()} WHERE release_id = ?")

    @classmethod
    def load_test_run(cls, run_id: UUID) -> 'SCTTestRun':
        return cls.get(id=run_id)

    @classmethod
    def submit_run(cls, request_data: dict) -> 'SCTTestRun':
        req = SCTTestRunSubmissionRequest(**request_data)
        return cls.from_sct_config(req=req)

    @classmethod
    def get_distinct_product_versions(cls, release: ArgusRelease) -> list[str]:
        cluster = ScyllaCluster.get()
        statement = cluster.prepare(f"SELECT scylla_version FROM {cls.table_name()} WHERE release_id = ?")
        rows = cluster.session.execute(query=statement, parameters=(release.id,))
        unique_versions = {r["scylla_version"] for r in rows if r["scylla_version"]}

        return sorted(list(unique_versions), reverse=True)

    @classmethod
    def get_version_data_for_release(cls, release_name: str):
        cluster = ScyllaCluster.get()
        release = ArgusRelease.get(name=release_name)
        query = cluster.prepare(f"SELECT scylla_version, packages, status FROM {cls.table_name()} WHERE release_id = ?")
        rows = cluster.session.execute(query=query, parameters=(release.id,))

        return list(rows)

    @classmethod
    def from_sct_config(cls, req: SCTTestRunSubmissionRequest):
        run = cls()
        run.build_id = req.job_name
        run.assign_categories()
        run.assignee = run.get_scheduled_assignee()
        run.start_time = datetime.utcnow()
        run.id = UUID(req.run_id)  # pylint: disable=invalid-name
        run.scm_revision_id = req.commit_id
        run.started_by = req.started_by
        run.build_job_url = req.job_url

        backend = req.sct_config.get("cluster_backend")
        region_key = SCT_REGION_PROPERTY_MAP.get(backend, SCT_REGION_PROPERTY_MAP["default"])
        raw_regions = req.sct_config.get(region_key) or "undefined_region"
        regions = raw_regions.split() if isinstance(raw_regions, str) else raw_regions
        primary_region = regions[0]

        run.cloud_setup = ResourceSetup.get_resource_setup(backend=backend, sct_config=req.sct_config)

        run.sct_runner_host = CloudInstanceDetails(
            public_ip=req.runner_public_ip,
            private_ip=req.runner_private_ip,
            provider=backend,
            region=primary_region,
        )

        run.config_files = req.sct_config.get("config_files")
        run.region_name = regions
        run.save()
        return run

    def get_resources(self) -> list[CloudResource]:
        return self.allocated_resources

    def get_nemeses(self) -> list[NemesisRunInfo]:
        return self.nemesis_data

    def get_events(self) -> list[EventsBySeverity]:
        return self.events

    def submit_product_version(self, version: str):
        self.scylla_version = version

    def finish_run(self):
        self.end_time = datetime.utcnow()

    def submit_logs(self, logs: list[dict]):
        for log in logs:
            self.logs.append((log["name"], log["link"]))

    def add_screenshot(self, screenshot_link: str):
        self.screenshots.append(screenshot_link)

    def add_nemesis(self, nemesis: NemesisRunInfo):
        self.nemesis_data.append(nemesis)

    def _add_new_event_type(self, event: EventsBySeverity):
        self.events.append(event)

    def _collect_event_message(self, event: EventsBySeverity, message: str):
        if len(event.last_events) >= 100:
            event.last_events = event.last_events[1:]

        event.event_amount += 1
        event.last_events.append(message)

    def add_event(self, event_severity: str, event_message: str):
        try:
            event = next(filter(lambda v: v.severity ==
                         event_severity, self.events))
        except StopIteration:
            event = EventsBySeverity(
                severity=event_severity, event_amount=0, last_events=[])
            self._add_new_event_type(event)

        self._collect_event_message(event, event_message)



# TODO: Deprecated, will be removed once REST API client is ready
from datetime import datetime
import re
import logging
import json
from uuid import UUID
from hashlib import sha1
from dataclasses import fields as dataclass_fields
from typing import KeysView, Union, Optional, Any, get_args as get_type_args, get_origin as get_type_origin
from types import GenericAlias

import cassandra.cluster
import cassandra.cqltypes
from cassandra import ConsistencyLevel
from cassandra.auth import PlainTextAuthProvider
from cassandra.query import named_tuple_factory
from cassandra.policies import WhiteListRoundRobinPolicy, AddressTranslator
from cassandra.cluster import ExecutionProfile, EXEC_PROFILE_DEFAULT
from cassandra.cqlengine import connection
from cassandra.cqlengine import models
from cassandra.cqlengine import management
from argus.db.argus_json import ArgusJSONEncoder

from argus.db.config import BaseConfig, FileConfig
from argus.db.db_types import ColumnInfo, CollectionHint, ArgusUDTBase
from argus.db.cloud_types import ResourceState
from argus.backend.models.web import ArgusSchedule, ArgusScheduleAssignee, \
    ArgusScheduleGroup, ArgusScheduleTest, ArgusRelease, ArgusGroup, ArgusTest

LOGGER = logging.getLogger(__name__)


class ArgusInterfaceSingletonError(Exception):
    pass


class ArgusInterfaceDatabaseConnectionError(Exception):
    pass


class ArgusInterfaceSchemaError(Exception):
    pass


class ArgusInterfaceNameError(Exception):
    pass


class PrivateToPublicAddressTranslator(AddressTranslator):
    def __init__(self, address_mapping):
        self.address_mapping = address_mapping

    def translate(self, addr):
        return self.address_mapping.get(addr, addr)


class ArgusDatabase:
    # pylint: disable=too-many-instance-attributes
    CQL_ENGINE_CONNECTION_NAME = 'argus_cql_engine_conn'
    ARGUS_EXECUTION_PROFILE = "argus_named_tuple"
    REQUIRED_CQL_ENGINE_MODELS = [
        ArgusSchedule,
        ArgusScheduleGroup,
        ArgusScheduleTest,
        ArgusScheduleAssignee,
        ArgusTest,
        ArgusGroup,
        ArgusRelease,
    ]
    PYTHON_SCYLLA_TYPE_MAPPING = {
        int: cassandra.cqltypes.IntegerType.typename,
        float: cassandra.cqltypes.FloatType.typename,
        str: cassandra.cqltypes.VarcharType.typename,
        UUID: cassandra.cqltypes.UUIDType.typename,
        datetime: cassandra.cqltypes.DateType.typename,
        ResourceState: cassandra.cqltypes.VarcharType.typename,
        Optional[str]: cassandra.cqltypes.VarcharType.typename,
        Optional[int]: cassandra.cqltypes.IntegerType.typename,
    }

    _INSTANCE: Union['ArgusDatabase', Any] = None

    def __init__(self, config: BaseConfig = None):
        if not config:
            config = FileConfig()
        self.config = config
        self.execution_profile = ExecutionProfile(
            load_balancing_policy=WhiteListRoundRobinPolicy(hosts=self.config.contact_points),
            consistency_level=ConsistencyLevel.QUORUM,
        )
        self.exec_profile_named_tuple = ExecutionProfile(
            load_balancing_policy=WhiteListRoundRobinPolicy(hosts=self.config.contact_points),
            consistency_level=ConsistencyLevel.QUORUM,
            row_factory=named_tuple_factory
        )
        address_translator = None
        if self.config.address_mapping:
            address_translator = PrivateToPublicAddressTranslator(self.config.address_mapping)

        self.cluster = cassandra.cluster.Cluster(contact_points=self.config.contact_points,
                                                 protocol_version=4,
                                                 auth_provider=PlainTextAuthProvider(
                                                     username=self.config.username,
                                                     password=self.config.password),
                                                 execution_profiles={
                                                     EXEC_PROFILE_DEFAULT: self.execution_profile,
                                                     self.ARGUS_EXECUTION_PROFILE: self.exec_profile_named_tuple
                                                 },
                                                 address_translator=address_translator,
                                                 )
        self.session = self.cluster.connect()
        self._keyspace_initialized = False
        self.prepared_statements = {}
        self.initialized_tables = {}
        self._table_keys = {}
        self._mapped_udts = {}
        self._current_keyspace = self.init_keyspace(name=self.config.keyspace_name)
        connection.register_connection(self.CQL_ENGINE_CONNECTION_NAME, session=self.session)
        for model in self.REQUIRED_CQL_ENGINE_MODELS:
            management.sync_table(model, keyspaces=(self._current_keyspace,),
                                  connections=(self.CQL_ENGINE_CONNECTION_NAME,))
        if not models.DEFAULT_KEYSPACE:
            models.DEFAULT_KEYSPACE = self._current_keyspace
        elif models.DEFAULT_KEYSPACE != self._current_keyspace:
            LOGGER.warning(
                "CQL Engine DEFAULT_KEYSPACE has been set already and differs from interface keyspace, this could cause issues")

    @classmethod
    def get(cls, config: BaseConfig = None):
        if cls._INSTANCE:
            LOGGER.debug("Found valid db session.")
            return cls._INSTANCE

        if not config:
            config = FileConfig()

        LOGGER.debug("Initializing db session from default config")
        cls._INSTANCE = cls(config=config)
        return cls._INSTANCE

    @classmethod
    def destroy(cls):
        if not cls._INSTANCE:
            LOGGER.warning("ArgusDatabase::destroy called with no valid session.")
            return False

        LOGGER.info("Shutting down the cluster connection.")
        cls._INSTANCE.cluster.shutdown()
        cls._INSTANCE = None
        return True

    @classmethod
    def from_config(cls, config: BaseConfig = None):
        return cls.get(config)

    def prepare_query_for_table(self, table_name, query_type, query):
        prepared_statement = self.session.prepare(query=query)
        self.prepared_statements[f"{table_name}_{query_type}"] = prepared_statement

        return prepared_statement

    @staticmethod
    def _verify_keyspace_name(name: str):
        incorrect_keyspace_name_re = r"\."
        if match := re.search(incorrect_keyspace_name_re, name):
            raise ArgusInterfaceNameError("Keyspace name does not conform to the "
                                          f"keyspace naming rules: {name} (pos: {match.pos})")
        return name

    @staticmethod
    def _get_hash_from_keys(keys: Union[list[str], KeysView]):
        key_string = ".".join(keys).encode(encoding="utf-8")
        return sha1(key_string).hexdigest()

    def init_keyspace(self, name="argus", prefix="", suffix="") -> str:
        keyspace_name = self._verify_keyspace_name(f"{prefix}{name}{suffix}")
        query = f"CREATE KEYSPACE IF NOT EXISTS {keyspace_name} " \
                "WITH replication={'class': 'SimpleStrategy', 'replication_factor' : 3}"
        LOGGER.debug("Running query: %s", query)
        self.session.execute(query=query, execution_profile=self.ARGUS_EXECUTION_PROFILE)
        self.session.set_keyspace(keyspace_name)
        self._keyspace_initialized = True
        return keyspace_name

    def is_native_type(self, object_type):
        return self.PYTHON_SCYLLA_TYPE_MAPPING.get(object_type, False)

    def init_table(self, table_name: str, column_info: dict[str, ColumnInfo]):
        # pylint: disable=too-many-locals
        if not self._keyspace_initialized:
            raise ArgusInterfaceDatabaseConnectionError("Uninitialized keyspace, cannot continue")

        if self.initialized_tables.get(table_name):
            return True, f"Table {table_name} already initialized"

        primary_keys_info: dict = column_info.pop("$tablekeys$")
        clustering_order_info: dict = column_info.pop("$clustering_order$")
        indices_info: dict = column_info.pop("$indices$")
        partition_keys = [key for key, (cls, pk_type) in primary_keys_info.items() if pk_type == "partition"]
        partition_key_def = partition_keys[0] if len(partition_keys) == 1 else f"({', '.join(partition_keys)})"
        clustering_columns = [key for key, (cls, pk_type) in primary_keys_info.items() if pk_type == "clustering"]
        clustering_column_def = ", ".join(clustering_columns)
        clustering_order = [f"{key} {clustering_order_info[key]}" for key in clustering_columns +
                            partition_keys if clustering_order_info.get(key)]
        clustering_order_def = ", ".join(clustering_order)
        clustering_statement = f"WITH CLUSTERING ORDER BY ({clustering_order_def})" if len(clustering_order) > 0 else ""

        self._table_keys[table_name] = primary_keys_info
        primary_key_def = f"{partition_key_def}, {clustering_column_def}" if len(
            clustering_column_def) > 0 else partition_key_def
        query = "CREATE TABLE IF NOT EXISTS {table_name}({columns}, PRIMARY KEY ({pk})) {cs}"
        columns_query = []
        for column in column_info.values():
            if mapped_type := self.is_native_type(column.type):
                column_type = mapped_type
            elif column.type is CollectionHint:
                column_type = self.create_collection_declaration(column.value.stored_type)
            else:
                # UDT
                column_type = f"frozen<{self._init_user_data_type(column.type)}>"

            constraints = " ".join(column.constraints)
            column_query = f"{column.name} {column_type} {constraints}"
            columns_query.append(column_query)

        columns_query = ", ".join(columns_query)
        completed_query = query.format(table_name=table_name, columns=columns_query,
                                       pk=primary_key_def, cs=clustering_statement)
        LOGGER.debug("About to execute: \"%s\"", completed_query)
        self.session.execute(query=completed_query, execution_profile=self.ARGUS_EXECUTION_PROFILE)
        self.create_indices(table_name, indices=indices_info)
        self.initialized_tables[table_name] = True
        return True, "Initialization complete"

    def create_indices(self, table_name, indices):
        for index in indices:
            self.session.execute(f"CREATE INDEX IF NOT EXISTS ON {table_name}({index})")
        return True

    def create_collection_declaration(self, hint: GenericAlias):
        collection_type = get_type_origin(hint)
        collection_types = get_type_args(hint)

        declaration_type = collection_type.__name__

        declared_types = []
        for inner_hint in collection_types:
            type_class = get_type_origin(inner_hint) if isinstance(inner_hint, GenericAlias) else inner_hint

            if type_class is tuple or type_class is list:
                declaration = f"frozen<{self.create_collection_declaration(inner_hint)}>"
            elif matched_type := self.PYTHON_SCYLLA_TYPE_MAPPING.get(type_class):
                declaration = matched_type
            else:
                declaration = f"frozen<{self._init_user_data_type(type_class)}>"

            declared_types.append(declaration)

        declaration_query = ", ".join(declared_types) if collection_type is tuple else str(declared_types[0])

        return f"{declaration_type}<{declaration_query}>"

    def _init_user_data_type(self, cls: ArgusUDTBase):
        if not self._keyspace_initialized:
            raise ArgusInterfaceDatabaseConnectionError("Uninitialized keyspace, cannot continue")

        udt_name = cls.basename()

        if cls in self._mapped_udts.get(self._current_keyspace, []):
            return udt_name

        query = "CREATE TYPE IF NOT EXISTS {name} ({fields})"
        fields = []
        for field in dataclass_fields(cls):
            name = field.name
            field_type = get_type_origin(field.type) if isinstance(field.type, GenericAlias) else field.type
            if field_type is list or field_type is tuple:
                field_declaration = self.create_collection_declaration(field.type)
            elif matched_type := self.PYTHON_SCYLLA_TYPE_MAPPING.get(field_type):
                field_declaration = matched_type
            else:
                field_declaration = f"frozen<{self._init_user_data_type(field.type)}>"
            fields.append(f"{name} {field_declaration}")

        joined_fields = ", ".join(fields)

        completed_query = query.format(name=udt_name, fields=joined_fields)
        LOGGER.debug("About to execute: \"%s\"", completed_query)
        self.session.execute(query=completed_query, execution_profile=self.ARGUS_EXECUTION_PROFILE)

        existing_udts = self._mapped_udts.get(self._current_keyspace, [])
        existing_udts.append(cls)
        self._mapped_udts[self._current_keyspace] = existing_udts

        return udt_name

    def fetch(self, table_name: str, run_id: UUID, where_clause="WHERE id = ?"):
        return self._fetch(table_name, (run_id,), where_clause)

    def _fetch(self, table_name: str, params: tuple | list, where_clause: str):
        if not self._keyspace_initialized:
            raise ArgusInterfaceDatabaseConnectionError("Uninitialized keyspace, cannot continue")

        query = self.prepared_statements.get(f"{table_name}_select_{where_clause.lower().replace(' ', '-')}",
                                             self.prepare_query_for_table(table_name=table_name, query_type="insert",
                                                                          query=f"SELECT * FROM {table_name} "
                                                                                f"{where_clause}"))

        cursor = self.session.execute(query=query, parameters=params, execution_profile=self.ARGUS_EXECUTION_PROFILE)

        return cursor.one()

    def fetch_generic(self, table_name: str, params: tuple | list, where_clause: str):
        return self._fetch(table_name, params, where_clause)

    def insert(self, table_name: str, run_data: dict):
        if not self._keyspace_initialized:
            raise ArgusInterfaceDatabaseConnectionError("Uninitialized keyspace, cannot continue")

        query = self.prepared_statements.get(f"{table_name}_insert",
                                             self.prepare_query_for_table(table_name=table_name, query_type="insert",
                                                                          query=f"INSERT INTO {table_name} JSON ?"))

        self.session.execute(query=query, parameters=(json.dumps(run_data, cls=ArgusJSONEncoder),),
                             execution_profile=self.ARGUS_EXECUTION_PROFILE)

    def update(self, table_name: str, run_data: dict):
        # pylint: disable=too-many-locals
        def _convert_data_to_sequence(data: dict) -> list:
            data_list = list(data.values())
            for idx, value in enumerate(data_list):
                if isinstance(value, dict):
                    data_list[idx] = _convert_data_to_sequence(value)
                elif isinstance(value, list) and len(value) > 0 and isinstance(value[0], dict):
                    data_list[idx] = [_convert_data_to_sequence(d) for d in value]
            return data_list

        primary_keys: dict = self._table_keys.get(table_name)
        if not primary_keys:
            raise ArgusInterfaceSchemaError(f"Table \"{table_name}\" is not initialized!")

        LOGGER.debug("Primary keys for table %s: %s", table_name, primary_keys)
        where_clause = []
        where_params = []
        for key, (_, _) in primary_keys.items():
            LOGGER.debug("Ejecting %s from update set as it is a part of the primary key", key)
            try:
                data_value = run_data.pop(key)
            except KeyError as exc:
                raise ArgusInterfaceSchemaError("Missing key from update set", key) from exc
            field = f"{key} = ?"
            where_clause.append(field)
            where_params.append(data_value)

        where_clause_joined = " AND ".join(where_clause)

        if not (prepared_statement := self.prepared_statements.get(f"update_{table_name}")):
            field_parameters = [f"\"{field_name}\" = ?" for field_name in run_data.keys()]
            fields_joined = ", ".join(field_parameters)

            query = f"UPDATE {table_name} SET {fields_joined} WHERE {where_clause_joined}"
            LOGGER.debug("Formatted query: %s", query)
            prepared_statement = self.session.prepare(query=query)
            self.prepared_statements[f"update_{table_name}"] = prepared_statement

        LOGGER.debug("Bound query for update: %s", prepared_statement.query_string)
        query_parameters = _convert_data_to_sequence(run_data)
        parameters = [*query_parameters, *where_params]
        LOGGER.debug("Parameters: %s", parameters)
        self.session.execute(prepared_statement, parameters=parameters, execution_profile=self.ARGUS_EXECUTION_PROFILE)



from uuid import UUID
from dataclasses import asdict
from argus.backend.util.enums import ResourceState, TestStatus
from argus.client.base import ArgusClient
from argus.client.sct.types import EventsInfo, LogLink, Package


class ArgusSCTClient(ArgusClient):
    test_type = "scylla-cluster-tests"
    schema_version: None = "v8"

    class Routes(ArgusClient.Routes):
        SUBMIT_PACKAGES = "/sct/$id/packages/submit"
        SUBMIT_SCREENSHOTS = "/sct/$id/screenshots/submit"
        CREATE_RESOURCE = "/sct/$id/resource/create"
        TERMINATE_RESOURCE = "/sct/$id/resource/$name/terminate"
        UPDATE_SHARDS_FOR_RESOURCE = "/sct/$id/resource/$name/shards"
        SUBMIT_NEMESIS = "/sct/$id/nemesis/submit"
        FINALIZE_NEMESIS = "/sct/$id/nemesis/finalize"
        SUBMIT_EVENTS = "/sct/$id/events/submit"

    def __init__(self, run_id: UUID, auth_token: str, base_url: str, api_version="v1") -> None:
        super().__init__(auth_token, base_url, api_version)
        self.run_id = run_id

    def submit_sct_run(self, job_name: str, job_url: str, started_by: str, commit_id: str,
                       runner_public_ip: str, runner_private_ip: str, sct_config: dict) -> None:
        """
            Submits an SCT run to argus.
        """
        response = super().submit_run(run_type=self.test_type, run_body={
            "run_id": str(self.run_id),
            "job_name": job_name,
            "job_url": job_url,
            "started_by": started_by,
            "commit_id": commit_id,
            "runner_public_ip": runner_public_ip,
            "runner_private_ip": runner_private_ip,
            "sct_config": sct_config,
        })

        self.check_response(response)

    def set_sct_run_status(self, new_status: TestStatus) -> None:
        """
            Sets an SCT run's status.
        """
        response = super().set_status(run_type=self.test_type, run_id=self.run_id, new_status=new_status)
        self.check_response(response)

    def update_scylla_version(self, version: str) -> None:
        """
            Updates scylla server version used for filtering test results by version.
        """
        response = super().update_product_version(run_type=self.test_type, run_id=self.run_id, product_version=version)
        self.check_response(response)

    def submit_sct_logs(self, logs: list[LogLink]) -> None:
        """
            Submits links to logs collected from nodes by SCT
        """
        response = super().submit_logs(run_type=self.test_type, run_id=self.run_id, logs=logs)
        self.check_response(response)

    def finalize_sct_run(self) -> None:
        """
            Marks run as finished inside argus. Currently this only sets end time of the run,
            run status must be updated separately by .set_sct_run_status
        """
        response = super().finalize_run(run_type=self.test_type, run_id=self.run_id)
        self.check_response(response)

    def submit_packages(self, packages: list[Package]):
        """
            Submits packages, which are relevant to this test - kernel versions, scylla-server versions, etc
        """
        response = self.post(
            endpoint=self.Routes.SUBMIT_PACKAGES,
            location_params={"id": str(self.run_id)},
            body={
                **self.generic_body,
                "packages": [asdict(p) for p in packages],
            }
        )
        self.check_response(response)

    def submit_screenshots(self, screenshot_links: list[str]) -> None:
        """
            Submits links to the screenshots from grafana, taken at the end of the test.
        """
        response = self.post(
            endpoint=self.Routes.SUBMIT_SCREENSHOTS,
            location_params={"id": str(self.run_id)},
            body={
                **self.generic_body,
                "screenshot_links": screenshot_links,
            }
        )
        self.check_response(response)

    def create_resource(self, name: str, resource_type: str, public_ip: str, private_ip: str,
                        region: str, provider: str, shards_amount: int, state=ResourceState.RUNNING) -> None:
        """
            Creates a cloud resource record in argus.
        """
        response = self.post(
            endpoint=self.Routes.CREATE_RESOURCE,
            location_params={"id": str(self.run_id)},
            body={
                **self.generic_body,
                "resource": {
                    "name": name,
                    "state": state,
                    "resource_type": resource_type,
                    "instance_details": {
                        "provider": provider,
                        "region": region,
                        "public_ip": public_ip,
                        "private_ip": private_ip,
                        "shards_amount": shards_amount,
                    }
                },
            }
        )
        self.check_response(response)

    def terminate_resource(self, name: str, reason: str) -> None:
        """
            Marks already created resource record as terminated.
        """
        response = self.post(
            endpoint=self.Routes.TERMINATE_RESOURCE,
            location_params={"id": str(self.run_id), "name": name},
            body={
                **self.generic_body,
                "reason": reason,
            }
        )
        self.check_response(response)

    def update_shards_for_resource(self, name: str, new_shards: int) -> None:
        """
            Separate methods to update "shards" of a resource, in cases
            where accurate reporting of shards is not possible during resource creation.
        """
        response = self.post(
            endpoint=self.Routes.UPDATE_SHARDS_FOR_RESOURCE,
            location_params={"id": str(self.run_id), "name": name},
            body={
                **self.generic_body,
                "shards": new_shards,
            }
        )
        self.check_response(response)

    def submit_nemesis(self, name: str, class_name: str, start_time: int,
                       target_name: str, target_ip: str, target_shards: int) -> None:
        """
            Submits a nemesis record. Should then be finalized by
            .finalize_nemesis method on nemesis completion.
        """
        response = self.post(
            endpoint=self.Routes.SUBMIT_NEMESIS,
            location_params={"id": str(self.run_id)},
            body={
                **self.generic_body,
                "nemesis": {
                    "name": name,
                    "class_name": class_name,
                    "start_time": start_time,
                    "node_name": target_name,
                    "node_ip": target_ip,
                    "node_shards": target_shards,
                }
            }
        )
        self.check_response(response)

    def finalize_nemesis(self, name: str, start_time: int, status: str, message: str) -> None:
        """
            Finalizes an already submitted nemesis. Sets the status, nemesis message and end time.
        """
        response = self.post(
            endpoint=self.Routes.FINALIZE_NEMESIS,
            location_params={"id": str(self.run_id)},
            body={
                **self.generic_body,
                "nemesis": {
                    "name": name,
                    "start_time": start_time,
                    "status": status,
                    "message": message,
                }
            }
        )
        self.check_response(response)

    def submit_events(self, events: list[EventsInfo]) -> None:
        """
            Submits events collected by event collector once at the end of the test.
        """
        response = self.post(
            endpoint=self.Routes.SUBMIT_EVENTS,
            location_params={"id": str(self.run_id)},
            body={
                **self.generic_body,
                "events": [asdict(e) for e in events]
            }
        )
        self.check_response(response)

    def sct_heartbeat(self) -> None:
        """
            Updates a heartbeat field for an already existing test.
        """
        super().heartbeat(run_type=self.test_type, run_id=self.run_id)



from abc import ABC
from typing import Hashable, Any
from pathlib import Path
from os import getenv
import logging
import yaml

LOGGER = logging.getLogger(__name__)


class ConfigLocationError(Exception):
    pass


class BaseConfig(ABC):
    def __init__(self):
        pass

    @property
    def as_dict(self) -> dict[Hashable, Any]:
        raise NotImplementedError()

    @property
    def username(self) -> str:
        raise NotImplementedError()

    @property
    def password(self) -> str:
        raise NotImplementedError()

    @property
    def contact_points(self) -> list[str]:
        raise NotImplementedError()

    @property
    def keyspace_name(self) -> str:
        raise NotImplementedError()

    @property
    def address_mapping(self) -> dict:
        return NotImplementedError()


class FileConfig(BaseConfig):
    @property
    def username(self):
        return self.as_dict.get("username")

    @property
    def password(self) -> str:
        return self.as_dict.get("password")

    @property
    def contact_points(self) -> list[str]:
        return self.as_dict.get("contact_points")

    @property
    def keyspace_name(self) -> str:
        return self.as_dict.get("keyspace_name")

    @property
    def address_mapping(self) -> dict:
        return self.as_dict.get("address_mapping")

    DEFAULT_CONFIG_PATHS = (
        "./config/argus.local.yaml",
        "argus.local.yaml",
        "argus.yaml",
        getenv("HOME") + "/.argus.yaml"
    )

    def __init__(self, filepath: str = None):
        super().__init__()
        if not filepath:
            for file in self.DEFAULT_CONFIG_PATHS:
                LOGGER.info("Trying %s", file)
                if Path(file).exists():
                    filepath = file
                    break

        if not filepath:
            LOGGER.error("All config locations were tried and no config file found")
            raise ConfigLocationError("No config file supplied and no config exists at default location")

        self.filepath = filepath
        self._credentials = None

    @property
    def as_dict(self) -> dict[Hashable, Any]:
        if self._credentials:
            return self._credentials
        path = Path(self.filepath)
        if not path.exists():
            raise ConfigLocationError(f"File not found: {self.filepath}")

        with open(path.absolute(), "rt", encoding="utf-8") as file:
            self._credentials = yaml.safe_load(file)

        return self._credentials


class Config(BaseConfig):
    @property
    def username(self) -> str:
        return self.as_dict.get("username")

    @property
    def password(self) -> str:
        return self.as_dict.get("password")

    @property
    def contact_points(self) -> list[str]:
        return self.as_dict.get("contact_points")

    @property
    def keyspace_name(self) -> str:
        return self.as_dict.get("keyspace_name")

    @property
    def address_mapping(self) -> dict:
        return self.as_dict.get("address_mapping")

    def __init__(self, username: str, password: str, contact_points: list[str], keyspace_name: str, address_mapping: dict | None = None):
        super().__init__()
        self._config = {
            "username": username,
            "password": password,
            "contact_points": contact_points,
            "keyspace_name": keyspace_name,
            "address_mapping": address_mapping,
        }

    @property
    def as_dict(self) -> dict[Hashable, Any]:
        return self._config



from uuid import UUID, uuid1, uuid4
from datetime import datetime
from enum import Enum, IntEnum, auto
from cassandra.cqlengine.models import Model
from cassandra.cqlengine.usertype import UserType
from cassandra.cqlengine import columns
from cassandra.util import uuid_from_time, unix_time_from_uuid1  # pylint: disable=no-name-in-module


def uuid_now():
    return uuid_from_time(datetime.utcnow())

# pylint: disable=invalid-name


class ArgusTestException(Exception):
    pass


class UserRoles(str, Enum):
    User = "ROLE_USER"
    Manager = "ROLE_MANAGER"
    Admin = "ROLE_ADMIN"


class User(Model):
    id = columns.UUID(primary_key=True, default=uuid4)
    username = columns.Text(index=True)
    full_name = columns.Text()
    password = columns.Text()
    email = columns.Text(index=True)
    registration_date = columns.DateTime()
    roles = columns.List(value_type=columns.Text)
    picture_id = columns.UUID(default=None)
    api_token = columns.Text(index=True)

    def __hash__(self) -> int:
        return hash(self.id)

    def is_manager(self) -> bool:
        # pylint: disable=unsupported-membership-test
        return UserRoles.Manager in self.roles

    def is_admin(self) -> bool:
        # pylint: disable=unsupported-membership-test
        return UserRoles.Admin in self.roles

    def set_as_admin(self) -> None:
        # pylint: disable=unsupported-membership-test
        if UserRoles.Admin not in self.roles:
            self.roles.append(UserRoles.Admin.value)

    def set_as_manager(self) -> None:
        # pylint: disable=unsupported-membership-test
        if UserRoles.Manager not in self.roles:
            self.roles.append(UserRoles.Manager.value)

    def get_id(self):
        return str(self.id)

    @classmethod
    def exists(cls, user_id: UUID):
        try:
            user = cls.get(id=user_id)
            if user:
                return user
        except cls.DoesNotExist:
            pass
        return None

    @classmethod
    def exists_by_name(cls, name: str):
        try:
            user = cls.get(username=name)
            if user:
                return user
        except cls.DoesNotExist:
            pass
        return None

    def __str__(self):
        return f"User('{self.id}','{self.username}')"

    def to_json(self):
        return {
            "id": str(self.id),
            "username": self.username,
            "full_name": self.full_name,
            "picture_id": self.picture_id
        }


class Team(Model):
    id = columns.UUID(primary_key=True, default=uuid4)
    name = columns.Text(required=True)
    leader = columns.UUID(index=True, required=True)
    members = columns.List(value_type=columns.UUID)
    motd = columns.Text()

class UserOauthToken(Model):
    id = columns.UUID(primary_key=True, default=uuid4)
    user_id = columns.UUID(index=True, required=True)
    kind = columns.Text(required=True, index=True)
    token = columns.Text(required=True)


class ArgusRelease(Model):
    __table_name__ = "argus_release_v2"
    id = columns.UUID(primary_key=True, default=uuid4)
    name = columns.Text(index=True, required=True)
    pretty_name = columns.Text()
    description = columns.Text()
    github_repo_url = columns.Text()
    assignee = columns.List(value_type=columns.UUID)
    picture_id = columns.UUID()
    enabled = columns.Boolean(default=lambda: True)
    perpetual = columns.Boolean(default=lambda: True)
    dormant = columns.Boolean(default=lambda: False)

    def __eq__(self, other):
        if isinstance(other, ArgusRelease):
            return self.name == other.name
        else:
            return super().__eq__(other)


class ArgusGroup(Model):
    __table_name__ = "argus_group_v2"
    id = columns.UUID(primary_key=True, default=uuid4)
    release_id = columns.UUID(required=True, index=True)
    name = columns.Text(required=True, index=True)
    pretty_name = columns.Text()
    description = columns.Text()
    assignee = columns.List(value_type=columns.UUID)
    build_system_id = columns.Text()
    enabled = columns.Boolean(default=lambda: True)

    def __hash__(self) -> int:
        return hash((self.id, self.release_id))

    def __eq__(self, other):
        if isinstance(other, ArgusGroup):
            return self.name == other.name and self.release_id == other.release_id
        else:
            return super().__eq__(other)


class ArgusTest(Model):
    __table_name__ = "argus_test_v2"
    id = columns.UUID(primary_key=True, default=uuid4)
    group_id = columns.UUID(required=True, index=True)
    release_id = columns.UUID(required=True, index=True)
    name = columns.Text(required=True, index=True)
    pretty_name = columns.Text()
    description = columns.Text()
    assignee = columns.List(value_type=columns.UUID)
    build_system_id = columns.Text(index=True)
    enabled = columns.Boolean(default=lambda: True)
    build_system_url = columns.Text()
    plugin_name = columns.Text()

    def __eq__(self, other):
        if isinstance(other, ArgusTest):
            return self.name == other.name and self.group_id == other.group_id and self.release_id == other.release_id
        else:
            return super().__eq__(other)

    def validate_build_system_id(self):
        try:
            t = ArgusTest.get(build_system_id=self.build_system_id)
            if t.id != self.id:
                raise ArgusTestException("Build Id is already used by another test", t.id, self.id)
        except ArgusTest.DoesNotExist:
            pass


class ArgusTestRunComment(Model):
    id = columns.UUID(primary_key=True, default=uuid4, partition_key=True)
    test_run_id = columns.UUID(required=True, index=True)
    user_id = columns.UUID(required=True, index=True)
    release_id = columns.UUID(required=True, index=True)
    posted_at = columns.Integer(
        required=True, clustering_order="desc", primary_key=True)
    message = columns.Text(min_length=1, max_length=65535)
    mentions = columns.List(value_type=columns.UUID, default=[])
    reactions = columns.Map(key_type=columns.Text, value_type=columns.Integer)


class ArgusEventTypes(str, Enum):
    AssigneeChanged = "ARGUS_ASSIGNEE_CHANGE"
    TestRunStatusChanged = "ARGUS_TEST_RUN_STATUS_CHANGE"
    TestRunInvestigationStatusChanged = "ARGUS_TEST_RUN_INVESTIGATION_STATUS_CHANGE"
    TestRunCommentPosted = "ARGUS_TEST_RUN_COMMENT_POSTED"
    TestRunCommentUpdated = "ARGUS_TEST_RUN_COMMENT_UPDATED"
    TestRunCommentDeleted = "ARGUS_TEST_RUN_COMMENT_DELETED"
    TestRunIssueAdded = "ARGUS_TEST_RUN_ISSUE_ADDED"
    TestRunIssueRemoved = "ARGUS_TEST_RUN_ISSUE_REMOVED"


class ArgusEvent(Model):
    id = columns.UUID(primary_key=True, default=uuid4, partition_key=True)
    release_id = columns.UUID(index=True)
    group_id = columns.UUID(index=True)
    test_id = columns.UUID(index=True)
    run_id = columns.UUID(index=True)
    user_id = columns.UUID(index=True)
    kind = columns.Text(required=True, index=True)
    body = columns.Text(required=True)
    created_at = columns.DateTime(required=True)


class ArgusNotificationTypes(str, Enum):
    Mention = "TYPE_MENTION"
    StatusChange = "TYPE_STATUS_CHANGE"
    AssigneeChange = "TYPE_ASSIGNEE_CHANGE"
    ScheduleChange = "TYPE_SCHEDULE_CHANGE"


class ArgusNotificationSourceTypes(str, Enum):
    TestRun = "TEST_RUN"
    Schedule = "SCHEDULE"
    Comment = "COMMENT"


class ArgusNotificationState(IntEnum):
    UNREAD = auto()
    READ = auto()


class ArgusNotification(Model):
    receiver = columns.UUID(primary_key=True, partition_key=True)
    id = columns.TimeUUID(primary_key=True, clustering_order="DESC", default=uuid_now)
    type = columns.Text(required=True)
    state = columns.SmallInt(required=True, default=lambda: ArgusNotificationState.UNREAD)
    sender = columns.UUID(required=True)
    source_type = columns.Text(required=True)
    source_id = columns.UUID(required=True)
    title = columns.Text(required=True, max_length=1024)
    content = columns.Text(required=True, max_length=65535)

    def to_dict_short_summary(self) -> dict:
        return {
            "receiver": self.receiver,
            "sender": self.sender,
            "id": self.id,
            "created": unix_time_from_uuid1(self.id) * 1000,
            "title": self.title,
            "state": self.state,
        }

    def to_dict(self) -> dict:
        return {
            "receiver": self.receiver,
            "sender": self.sender,
            "id": self.id,
            "created": unix_time_from_uuid1(self.id) * 1000,
            "title": self.title,
            "type": self.type,
            "content": self.content,
            "source": self.source_type,
            "source_id": self.source_id,
            "state": self.state,
        }


class ArgusGithubIssue(Model):
    # pylint: disable=too-many-instance-attributes
    id = columns.UUID(primary_key=True, default=uuid4, partition_key=True)
    added_on = columns.DateTime(default=datetime.utcnow)
    release_id = columns.UUID(index=True)
    group_id = columns.UUID(index=True)
    test_id = columns.UUID(index=True)
    run_id = columns.UUID(index=True)
    user_id = columns.UUID(index=True)
    type = columns.Text()
    owner = columns.Text()
    repo = columns.Text()
    issue_number = columns.Integer()
    last_status = columns.Text()
    title = columns.Text()
    url = columns.Text()

    def __hash__(self) -> int:
        return hash((self.owner, self.repo, self.issue_number))

    def __eq__(self, other):
        if isinstance(other, ArgusGithubIssue):
            return self.owner == other.owner and self.repo == other.repo and self.issue_number == other.issue_number
        return super().__eq__(other)

    def __ne__(self, other):
        return not self == other


class ArgusSchedule(Model):
    __table_name__ = "argus_schedule_v4"
    release_id = columns.UUID(primary_key=True, required=True)
    id = columns.TimeUUID(primary_key=True, default=uuid1, clustering_order="DESC")
    period_start = columns.DateTime(required=True, default=datetime.utcnow)
    period_end = columns.DateTime(required=True, primary_key=True, clustering_order="DESC")
    tag = columns.Text(default="")


class ArgusScheduleAssignee(Model):
    __table_name__ = "argus_schedule_user_v3"
    assignee = columns.UUID(primary_key=True)
    id = columns.TimeUUID(primary_key=True, default=uuid1,
                          clustering_order="DESC")
    schedule_id = columns.TimeUUID(required=True, index=True)
    release_id = columns.UUID(required=True)


class ArgusScheduleTest(Model):
    __table_name__ = "argus_schedule_test_v5"
    test_id = columns.UUID(primary_key=True, required=True)
    id = columns.TimeUUID(primary_key=True, default=uuid1,
                          clustering_order="DESC")
    schedule_id = columns.TimeUUID(required=True, index=True)
    release_id = columns.UUID(partition_key=True)


class ArgusScheduleGroup(Model):
    __table_name__ = "argus_schedule_group_v3"
    group_id = columns.UUID(partition_key=True, required=True)
    id = columns.TimeUUID(primary_key=True, default=uuid1,
                          clustering_order="DESC")
    schedule_id = columns.TimeUUID(required=True, index=True)
    release_id = columns.UUID(partition_key=True)


class ReleasePlannerComment(Model):
    __table_name__ = "argus_planner_comment_v2"
    release = columns.UUID(primary_key=True)
    group = columns.UUID(primary_key=True)
    test = columns.UUID(primary_key=True)
    comment = columns.Text(default=lambda: "")


class WebFileStorage(Model):
    id = columns.UUID(primary_key=True, default=uuid4)
    filepath = columns.Text(min_length=1)
    filename = columns.Text(min_length=1)


USED_MODELS: list[Model] = [
    User,
    UserOauthToken,
    Team,
    WebFileStorage,
    ArgusRelease,
    ArgusGroup,
    ArgusTest,
    ArgusTestRunComment,
    ArgusEvent,
    ArgusGithubIssue,
    ReleasePlannerComment,
    ArgusNotification,
    ArgusSchedule,
    ArgusScheduleAssignee,
    ArgusScheduleGroup,
    ArgusScheduleTest,
]

USED_TYPES: list[UserType] = [

]



import os
import hashlib
from flask import (
    Blueprint, flash, g, redirect, render_template, request, session, url_for, current_app
)
from werkzeug.security import check_password_hash
from argus.backend.models.web import User
from argus.backend.service.user import UserService, load_logged_in_user, login_required

bp = Blueprint('auth', __name__, url_prefix='/auth')


@bp.route('/register', methods=('GET', 'POST'))
def register():
    return redirect(url_for("auth.login"))


@bp.route('/login', methods=('GET', 'POST'))
def login():
    token = hashlib.sha256((os.urandom(64))).hexdigest()
    session["csrf_token"] = token

    if request.method == 'POST':
        username = request.form["username"]
        password = request.form["password"]
        error = None
        try:
            user = User.get(username=username)
        except User.DoesNotExist:
            user = None

        if not user:
            error = "User not found"
        elif not check_password_hash(user.password, password):
            error = "Incorrect Password"

        if not error:
            session.clear()
            session["user_id"] = str(user.id)
            session["csrf_token"] = token
        flash(error, category="error")
        return redirect(url_for('main.home'))

    return render_template('auth/login.html.j2',
                           csrf_token=token,
                           github_cid=current_app.config.get("GITHUB_CLIENT_ID", "NO_CLIENT_ID"),
                           github_scopes=current_app.config.get("GITHUB_SCOPES", "user:email read:user read:org repo")
                           )


@bp.route("/profile/api/token/generate", methods=("POST",))
@login_required
def generate_api_token():
    new_token = UserService().generate_token(g.user)
    session["token_generated"] = new_token
    return redirect(url_for('main.profile'))


@bp.route('/logout', methods=("POST",))
def logout():
    session.clear()
    return redirect(url_for('main.home'))


bp.before_app_request(load_logged_in_user)



from flask import Blueprint, request

from argus.backend.error_handlers import handle_api_exception
from argus.backend.service.user import api_login_required
from argus.backend.service.client_service import ClientService
from argus.backend.util.common import get_payload
from argus.backend.plugins.loader import AVAILABLE_PLUGINS

bp = Blueprint("client_api", __name__, url_prefix="/client")
bp.register_error_handler(Exception, handle_api_exception)
for plugin in AVAILABLE_PLUGINS.values():
    if plugin.controller:
        bp.register_blueprint(plugin.controller)


@bp.route("/testrun/<string:run_type>/submit", methods=["POST"])
@api_login_required
def submit_run(run_type: str):
    payload = get_payload(request)
    result = ClientService().submit_run(run_type=run_type, request_data=payload)
    return {
        "status": "ok",
        "response": result
    }


@bp.route("/testrun/<string:run_type>/<string:run_id>/heartbeat", methods=["POST"])
@api_login_required
def run_heartbeat(run_type: str, run_id: str):
    result = ClientService().heartbeat(run_type=run_type, run_id=run_id)
    return {
        "status": "ok",
        "response": result
    }


@bp.route("/testrun/<string:run_type>/<string:run_id>/set_status", methods=["POST"])
@api_login_required
def run_set_status(run_type: str, run_id: str):
    payload = get_payload(request)
    result = ClientService().update_run_status(run_type=run_type, run_id=run_id, new_status=payload["new_status"])
    return {
        "status": "ok",
        "response": result
    }


@bp.route("/testrun/<string:run_type>/<string:run_id>/update_product_version", methods=["POST"])
@api_login_required
def run_update_product_version(run_type: str, run_id: str):
    payload = get_payload(request)
    result = ClientService().submit_product_version(
        run_type=run_type, run_id=run_id, version=payload["product_version"])
    return {
        "status": "ok",
        "response": result
    }


@bp.route("/testrun/<string:run_type>/<string:run_id>/logs/submit", methods=["POST"])
@api_login_required
def run_submit_logs(run_type: str, run_id: str):
    payload = get_payload(request)
    result = ClientService().submit_logs(run_type=run_type, run_id=run_id, logs=payload["logs"])
    return {
        "status": "ok",
        "response": result
    }


@bp.route("/testrun/<string:run_type>/<string:run_id>/finalize", methods=["POST"])
@api_login_required
def run_finalize(run_type: str, run_id: str):
    result = ClientService().finish_run(run_type=run_type, run_id=run_id)
    return {
        "status": "ok",
        "response": result
    }



from typing import TypedDict


class RawMatrixTestCase(TypedDict):
    name: str
    status: str
    time: float
    classname: str
    message: str


class RawMatrixTestSuite(TypedDict):
    name: str
    tests: int
    failures: int
    disabled: int
    skipped: int
    passed: int
    errors: int
    time: float
    cases: list[RawMatrixTestCase]


class RawMatrixTestResult(TypedDict):
    name: str
    driver_name: str
    tests: int
    failures: int
    errors: int
    disabled: int
    skipped: int
    passed: int
    time: float
    timestamp: str
    suites: list[RawMatrixTestSuite]



import logging
from pathlib import Path

import yaml


LOGGER = logging.getLogger(__name__)


class Config:
    CONFIG = None
    CONFIG_PATHS = [
        Path("./config/argus_web.yaml"),
        Path("argus_web.yaml"),
    ]

    @classmethod
    def locate_argus_web_config(cls) -> Path:
        for config in cls.CONFIG_PATHS:
            if config.exists():
                return config
            else:
                LOGGER.debug("Tried %s as config, not found.", config)

        raise Exception("Failed to locate web application config file!")

    @classmethod
    def load_yaml_config(cls) -> dict:
        if cls.CONFIG:
            return cls.CONFIG
        path = cls.locate_argus_web_config()
        with open(path, "rt", encoding="utf-8") as file:
            config = yaml.safe_load(file)

        cls.CONFIG = config
        return config



from functools import cached_property
import logging
from typing import Optional
from flask import g, Flask
# pylint: disable=no-name-in-module
from cassandra.policies import WhiteListRoundRobinPolicy
from cassandra import ConsistencyLevel
from cassandra.cluster import ExecutionProfile, EXEC_PROFILE_DEFAULT, Cluster
from cassandra.cluster import PreparedStatement
from cassandra.cqlengine.management import sync_table, sync_type
from cassandra.cqlengine import connection
from cassandra.query import dict_factory
from cassandra.auth import PlainTextAuthProvider
from argus.backend.util.config import Config

from argus.backend.models.web import USED_MODELS, USED_TYPES

LOGGER = logging.getLogger(__name__)


class ScyllaCluster:
    # pylint: disable=too-many-instance-attributes
    APP_INSTANCE: Optional['ScyllaCluster'] = None

    def __init__(self, config=None):
        if not config:
            config = Config.load_yaml_config()
        self.config = config
        self.auth_provider = PlainTextAuthProvider(
            username=config["SCYLLA_USERNAME"], password=config["SCYLLA_PASSWORD"])
        self.lb_policy = WhiteListRoundRobinPolicy(hosts=config["SCYLLA_CONTACT_POINTS"])
        self.execution_profile = ExecutionProfile(
            load_balancing_policy=self.lb_policy, consistency_level=ConsistencyLevel.QUORUM)
        connection.setup(hosts=config["SCYLLA_CONTACT_POINTS"], default_keyspace=config["SCYLLA_KEYSPACE_NAME"],
                         auth_provider=self.auth_provider,
                         protocol_version=4,
                         execution_profiles={EXEC_PROFILE_DEFAULT: self.execution_profile})
        self.cluster: Cluster = connection.get_cluster(connection='default')
        self.prepared_statements = {}
        self.read_exec_profile = ExecutionProfile(
            consistency_level=ConsistencyLevel.ONE,
            row_factory=dict_factory,
            load_balancing_policy=self.lb_policy
        )
        self.read_named_tuple_exec_profile = ExecutionProfile(
            consistency_level=ConsistencyLevel.ONE,
            load_balancing_policy=self.lb_policy
        )
        self.cluster.add_execution_profile("read_fast", self.read_exec_profile)
        self.cluster.add_execution_profile("read_fast_named_tuple", self.read_named_tuple_exec_profile)

    @cached_property
    def session(self):
        return self.cluster.connect(keyspace=self.config["SCYLLA_KEYSPACE_NAME"])

    @classmethod
    def get(cls, config: Config = None) -> 'ScyllaCluster':
        if cls.APP_INSTANCE:
            return cls.APP_INSTANCE

        cls.APP_INSTANCE = cls(config)
        return cls.APP_INSTANCE

    @classmethod
    def shutdown(cls):
        if cls.APP_INSTANCE:
            cls.APP_INSTANCE.cluster.shutdown()
            cls.APP_INSTANCE = None

    def prepare(self, query: str) -> PreparedStatement:
        if not (statement := self.prepared_statements.get(query)):
            LOGGER.info("Unprepared statement %s, preparing...", query)
            statement = self.session.prepare(query=query)
            self.prepared_statements[query] = statement
        return statement

    def sync_core_tables(self):
        for udt in USED_TYPES:
            LOGGER.info("Syncing type: %s..", udt.__name__)
            sync_type(ks_name=self.config["SCYLLA_KEYSPACE_NAME"], type_model=udt)
        LOGGER.info("Core Types synchronized.")
        for model in USED_MODELS:
            LOGGER.info("Syncing model: %s..", model.__name__)
            sync_table(model, keyspaces=[self.config["SCYLLA_KEYSPACE_NAME"]])
        LOGGER.info("Core Models synchronized.")

    @classmethod
    def get_session(cls):
        cluster = cls.get()
        if 'scylla_session' not in g:
            g.scylla_session = cluster.session  # pylint: disable=assigning-non-slot
        return g.scylla_session

    @classmethod
    def close_session(cls, error=None):  # pylint: disable=unused-argument
        g.pop("scylla_session", None)

    @classmethod
    def attach_to_app(cls, app: Flask):
        app.teardown_appcontext(cls.close_session)



from datetime import datetime
from json.encoder import JSONEncoder
from uuid import UUID

import cassandra.cqlengine.usertype as ut
import cassandra.cqlengine.models as m


class ArgusJSONEncoder(JSONEncoder):
    def default(self, o):
        match o:
            case UUID():
                return str(o)
            case ut.UserType():
                return dict(o.items())
            case m.Model():
                return dict(o.items())
            case datetime():
                return o.strftime("%Y-%m-%d %H:%M:%S UTC")
            case _:
                return super().default(o)



import logging
from uuid import UUID
import requests
from flask import (
    Blueprint,
    g,
    request
)
from flask.json import jsonify
from argus.backend.error_handlers import handle_api_exception
from argus.backend.controller.notification_api import bp as notifications_bp
from argus.backend.controller.client_api import bp as client_bp
from argus.backend.controller.testrun_api import bp as testrun_bp
from argus.backend.controller.team import bp as team_bp
from argus.backend.service.argus_service import ArgusService
from argus.backend.service.user import UserService, api_login_required
from argus.backend.service.stats import ReleaseStatsCollector
from argus.backend.models.web import ArgusRelease, ArgusGroup, ArgusTest, User, UserOauthToken
from argus.backend.util.common import get_payload

bp = Blueprint('api', __name__, url_prefix='/api/v1')
bp.register_blueprint(notifications_bp)
bp.register_blueprint(client_bp)
bp.register_blueprint(testrun_bp)
bp.register_blueprint(team_bp)
bp.register_error_handler(Exception, handle_api_exception)
LOGGER = logging.getLogger(__name__)


@bp.route("/version")
def app_version():
    service = ArgusService()
    argus_version = service.get_version()
    return jsonify({
        "status": "ok",
        "response": {
            "commit_id": argus_version
        }
    })


@bp.route("/profile/github/token")
@api_login_required
def get_github_oauth_token():
    user_tokens = UserOauthToken.filter(user_id=g.user.id).all()
    token = None
    for tok in user_tokens:
        if tok.kind == "github":
            token = tok.token
            break
    if not token:
        raise Exception("Github token not found")

    res = jsonify({
        "status": "ok",
        "response": token
    })
    res.cache_control.max_age = 300

    return res


@bp.route("/releases")
@api_login_required
def releases():
    service = ArgusService()
    force_all = request.args.get("all", False)
    all_releases = service.get_releases()
    response = jsonify({
        "status": "ok",
        "response": [dict(d.items()) for d in all_releases if d.enabled or force_all]
    })

    response.cache_control.max_age = 60
    return response


@bp.route("/release/activity", methods=["GET"])
@api_login_required
def release_activity():
    release_name = request.args.get("releaseName")
    if not release_name:
        raise Exception("Release name not specified in the request")
    service = ArgusService()
    activity_data = service.fetch_release_activity(release_name)

    return jsonify({
        "status": "ok",
        "response": activity_data
    })


@bp.route("/release/planner/data", methods=["GET"])
@api_login_required
def release_planner_data():

    release_id = request.args.get("releaseId")
    if not release_id:
        raise Exception("Release Id not specified")
    service = ArgusService()
    planner_data = service.get_planner_data(release_id)
    return jsonify({
        "status": "ok",
        "response": planner_data
    })


@bp.route("/release/<string:release_id>/versions")
@api_login_required
def release_versions(release_id: str):
    release_id = UUID(release_id)
    service = ArgusService()
    distinct_versions = service.get_distinct_release_versions(release_id=release_id)

    return jsonify({
        "status": "ok",
        "response": distinct_versions
    })


@bp.route("/release/planner/comment/get/test")
def get_planner_comment_by_test():
    test_id = request.args.get("id")
    if not test_id:
        raise Exception("TestId was not specified")
    service = ArgusService()
    planner_comments_by_test = service.get_planner_comment_by_test(UUID(test_id))

    response = jsonify({
        "status": "ok",
        "response": planner_comments_by_test
    })
    response.cache_control.max_age = 60
    return response


@bp.route("/release/schedules/comment/update", methods=["POST"])
@api_login_required
def release_schedules_comment_update():
    if not request.is_json:
        raise Exception(
            "Content-Type mismatch, expected application/json, got:", request.content_type)
    request_payload = request.get_json()
    service = ArgusService()
    comment_update_result = service.update_schedule_comment(request_payload)

    return jsonify({
        "status": "ok",
        "response": comment_update_result
    })


@bp.route("/release/schedules", methods=["GET"])
@api_login_required
def release_schedules():
    release = request.args.get("releaseId")
    if not release:
        raise Exception("No releaseId provided")
    service = ArgusService()
    release_schedules_data = service.get_schedules_for_release(release)

    return jsonify({
        "status": "ok",
        "response": release_schedules_data
    })


@bp.route("/release/schedules/assignee/update", methods=["POST"])
@api_login_required
def release_schedules_assignee_update():
    if not request.is_json:
        raise Exception(
            "Content-Type mismatch, expected application/json, got:", request.content_type)
    request_payload = request.get_json()
    service = ArgusService()
    assignee_update_status = service.update_schedule_assignees(request_payload)

    return jsonify({
        "status": "ok",
        "response": assignee_update_status
    })


@bp.route("/release/assignees/groups", methods=["GET"])
@api_login_required
def group_assignees():
    release_id = request.args.get("releaseId")
    if not release_id:
        raise Exception("Missing releaseId")
    service = ArgusService()
    group_assignees_list = service.get_groups_assignees(release_id)

    return jsonify({
        "status": "ok",
        "response": group_assignees_list
    })


@bp.route("/release/assignees/tests", methods=["GET"])
@api_login_required
def tests_assignees():
    group_id = request.args.get("groupId")
    if not group_id:
        raise Exception("Missing groupId")
    service = ArgusService()
    tests_assignees_list = service.get_tests_assignees(group_id)

    return jsonify({
        "status": "ok",
        "response": tests_assignees_list
    })


@bp.route("/release/schedules/submit", methods=["POST"])
@api_login_required
def release_schedules_submit():
    if not request.is_json:
        raise Exception(
            "Content-Type mismatch, expected application/json, got:", request.content_type)
    payload = request.get_json()
    service = ArgusService()
    schedule_submit_result = service.submit_new_schedule(
        release=payload["releaseId"],
        start_time=payload["start"],
        end_time=payload["end"],
        tests=payload["tests"],
        groups=payload["groups"],
        assignees=payload["assignees"],
        tag=payload["tag"],
    )

    return jsonify({
        "status": "ok",
        "response": schedule_submit_result
    })


@bp.route("/release/schedules/delete", methods=["POST"])
@api_login_required
def release_schedules_delete():
    if not request.is_json:
        raise Exception(
            "Content-Type mismatch, expected application/json, got:", request.content_type)
    request_payload = request.get_json()
    service = ArgusService()
    schedule_delete_result = service.delete_schedule(request_payload)

    return jsonify({
        "status": "ok",
        "response": schedule_delete_result
    })


@bp.route("/groups", methods=["GET"])
@api_login_required
def argus_groups():
    release_id = request.args.get("releaseId")
    if not release_id:
        raise Exception("No releaseId provided")

    force_all = request.args.get("all", False)
    service = ArgusService()
    groups = service.get_groups(UUID(release_id))
    result_groups = [dict(g.items()) for g in groups if g.enabled or force_all]

    response = jsonify({
        "status": "ok",
        "response": result_groups
    })
    response.cache_control.max_age = 60
    return response


@bp.route("/tests", methods=["GET"])
@api_login_required
def argus_tests():
    group_id = request.args.get("groupId")
    if not group_id:
        raise Exception("No groupId provided")
    force_all = request.args.get("all", False)
    service = ArgusService()
    tests = service.get_tests(group_id=group_id)
    result_tests = [dict(t.items()) for t in tests if t.enabled or force_all]

    response = jsonify({
        "status": "ok",
        "response": result_tests
    })
    response.cache_control.max_age = 60
    return response


@bp.route("/release/<string:release_id>/details", methods=["GET"])
@api_login_required
def get_release_details(release_id: str):
    release = ArgusRelease.get(id=UUID(release_id))
    response = jsonify({
        "status": "ok",
        "response": release,
    })
    response.cache_control.max_age = 60
    return response


@bp.route("/group/<string:group_id>/details", methods=["GET"])
@api_login_required
def get_group_details(group_id: str):
    group = ArgusGroup.get(id=UUID(group_id))
    response = jsonify({
        "status": "ok",
        "response": group,
    })
    response.cache_control.max_age = 60
    return response


@bp.route("/test/<string:test_id>/details", methods=["GET"])
@api_login_required
def get_test_details(test_id: str):
    test = ArgusTest.get(id=UUID(test_id))
    response = jsonify({
        "status": "ok",
        "response": test
    })
    response.cache_control.max_age = 60
    return response


@bp.route("/test/<string:test_id>/set_plugin", methods=["POST"])
@api_login_required
def set_test_plugin(test_id: str):
    payload = get_payload(request)

    current_user: User = g.user
    if not current_user.is_admin():
        raise Exception("Cannot set plugin as a non-administrator")
    test: ArgusTest = ArgusTest.get(id=UUID(test_id))
    test.plugin_name = payload["plugin_name"]
    test.save()

    return {
        "status": "ok",
        "response": test
    }


@bp.route("/test-info", methods=["GET"])
@api_login_required
def test_info():
    test_id = request.args.get("testId")
    if not test_id:
        raise Exception("No testId provided")
    service = ArgusService()
    info = service.get_test_info(test_id=UUID(test_id))

    return {
        "status": "ok",
        "response": info
    }


@bp.route("/test_run/comment/get", methods=["GET"])  # TODO: remove
@api_login_required
def get_test_run_comment():
    comment_id = request.args.get("commentId")
    if not comment_id:
        raise Exception("commentId wasn't specified in the request")
    service = ArgusService()
    comment = service.get_comment(comment_id=UUID(comment_id))
    return jsonify({
        "status": "ok",
        "response": comment if comment else False
    })


@bp.route("/users", methods=["GET"])
@api_login_required
def user_info():
    result = UserService().get_users()

    return jsonify({
        "status": "ok",
        "response": result
    })


@bp.route("/release/stats/v2", methods=["GET"])
@api_login_required
def release_stats_v2():
    request.query_string.decode(encoding="UTF-8")
    release = request.args.get("release")
    limited = bool(int(request.args.get("limited")))
    version = request.args.get("productVersion", None)
    force = bool(int(request.args.get("force")))
    stats = ReleaseStatsCollector(
        release_name=release, release_version=version).collect(limited=limited, force=force)

    res = jsonify({
        "status": "ok",
        "response": stats
    })
    res.cache_control.max_age = 300
    return res


@bp.route("/test_runs/poll", methods=["GET"])
@api_login_required
def test_runs_poll():
    limit = request.args.get("limit")
    if not limit:
        limit = 10
    else:
        limit = int(limit)
    test_id = UUID(request.args.get('testId'))
    additional_runs = [UUID(run) for run in request.args.getlist('additionalRuns[]')]
    service = ArgusService()
    runs = service.poll_test_runs(test_id=test_id, additional_runs=additional_runs, limit=limit)

    return jsonify({
        "status": "ok",
        "response": runs
    })


@bp.route("/test_run/poll", methods=["GET"])
@api_login_required
def test_run_poll_single():
    runs = request.args.get("runs", "")
    runs = [UUID(r) for r in runs.split(",") if r]
    service = ArgusService()
    test_runs = service.poll_test_runs_single(runs=runs)

    return jsonify({
        "status": "ok",
        "response": test_runs
    })


@bp.route("/release/create", methods=["POST"])
@api_login_required
def release_create():
    if not request.is_json:
        raise Exception(
            "Content-Type mismatch, expected application/json, got:", request.content_type)
    request_payload = request.get_json()
    service = ArgusService()
    result = service.create_release(request_payload)

    return jsonify({
        "status": "ok",
        "response": result
    })


@bp.route("/artifact/resolveSize")
@api_login_required
def resolve_artifact_size():
    link = request.args.get("l")
    if not link:
        raise Exception("No link provided")

    res = requests.head(link)

    if res.status_code != 200:
        raise Exception("Error requesting resource")

    length = res.headers.get("Content-Length")
    if length:
        length = int(length)

    return {
        "status": "ok",
        "response": {
            "artifactSize": length,
        }
    }



import re
import logging
from dataclasses import asdict
from typing import Any, Type
from uuid import UUID

import requests

from argus.backend.util.enums import TestStatus
from argus.client.sct.types import LogLink

JSON = dict[str, Any] | list[Any] | int | str | float | bool | Type[None]
LOGGER = logging.getLogger(__name__)


class ArgusClientError(Exception):
    pass


class ArgusClient:
    schema_version: str | None = None

    class Routes():
        # pylint: disable=too-few-public-methods
        SUBMIT = "/testrun/$type/submit"
        HEARTBEAT = "/testrun/$type/$id/heartbeat"
        SET_STATUS = "/testrun/$type/$id/set_status"
        SET_PRODUCT_VERSION = "/testrun/$type/$id/update_product_version"
        SUBMIT_LOGS = "/testrun/$type/$id/logs/submit"
        FINALIZE = "/testrun/$type/$id/finalize"

    def __init__(self, auth_token: str, base_url: str, api_version="v1") -> None:
        self._auth_token = auth_token
        self._base_url = base_url
        self._api_ver = api_version

    @property
    def auth_token(self) -> str:
        return self._auth_token

    def verify_location_params(self, endpoint: str, location_params: dict[str, str]) -> bool:
        required_params: list[str] = re.findall(r"\$[\w_]+", endpoint)
        for param in required_params:
            if param.lstrip("$") not in location_params.keys():
                raise ArgusClientError(f"Missing required location argument for endpoint {endpoint}: {param}")

        return True

    @staticmethod
    def check_response(response: requests.Response, expected_code: int = 200):
        if response.status_code != expected_code:
            raise ArgusClientError(
                f"Unexpected HTTP Response encountered - expected: {expected_code}, got: {response.status_code}",
                expected_code,
                response.status_code,
                response.request,
            )

        response_data: JSON = response.json()
        LOGGER.debug("API Response: %s", response_data)
        if response_data.get("status") != "ok":
            raise ArgusClientError(
                f"API Error encountered using endpoint: {response.request.method} {response.request.path_url}",
                response_data["response"]["arguments"][0],
            )

    def get_url_for_endpoint(self, endpoint: str, location_params: dict[str, str] | None) -> str:
        if self.verify_location_params(endpoint, location_params):
            for param, value in location_params.items():
                endpoint = endpoint.replace(f"${param}", str(value))
        return f"{self._base_url}/api/{self._api_ver}/client{endpoint}"

    @property
    def generic_body(self) -> dict:
        return {
            "schema_version": self.schema_version
        }

    @property
    def request_headers(self):
        return {
            "Authorization": f"token {self.auth_token}",
            "Accept": "application/json",
            "Content-Type": "application/json",
        }

    def get(self, endpoint: str, location_params: dict[str, str] = None, params: dict = None) -> requests.Response:
        response = requests.get(
            url=self.get_url_for_endpoint(
                endpoint=endpoint,
                location_params=location_params
            ),
            params=params,
            headers=self.request_headers
        )

        return response

    def post(
        self,
        endpoint: str,
        location_params: dict = None,
        params: dict = None,
        body: dict = None,
    ) -> requests.Response:
        response = requests.post(
            url=self.get_url_for_endpoint(
                endpoint=endpoint,
                location_params=location_params
            ),
            params=params,
            json=body,
            headers=self.request_headers,
        )

        return response

    def submit_run(self, run_type: str, run_body: dict) -> requests.Response:
        return self.post(endpoint=self.Routes.SUBMIT, location_params={"type": run_type}, body={
            **self.generic_body,
            **run_body
        })

    def set_status(self, run_type: str, run_id: UUID, new_status: TestStatus) -> requests.Response:
        return self.post(
            endpoint=self.Routes.SET_STATUS,
            location_params={"type": run_type, "id": str(run_id)},
            body={
                **self.generic_body,
                "new_status": new_status.value
            }
        )

    def update_product_version(self, run_type: str, run_id: UUID, product_version: str) -> requests.Response:
        return self.post(
            endpoint=self.Routes.SET_PRODUCT_VERSION,
            location_params={"type": run_type, "id": str(run_id)},
            body={
                **self.generic_body,
                "product_version": product_version
            }
        )

    def submit_logs(self, run_type: str, run_id: UUID, logs: list[LogLink]) -> requests.Response:
        return self.post(
            endpoint=self.Routes.SUBMIT_LOGS,
            location_params={"type": run_type, "id": str(run_id)},
            body={
                **self.generic_body,
                "logs": [asdict(l) for l in logs]
            }
        )

    def finalize_run(self, run_type: str, run_id: UUID) -> requests.Response:
        return self.post(
            endpoint=self.Routes.FINALIZE,
            location_params={"type": run_type, "id": str(run_id)},
            body={
                **self.generic_body,
            }
        )

    def heartbeat(self, run_type: str, run_id: UUID) -> None:
        response = self.post(
            endpoint=self.Routes.HEARTBEAT,
            location_params={"type": run_type, "id": str(run_id)},
            body={
                **self.generic_body,
            }
        )
        self.check_response(response)



import sys
from functools import wraps
from typing import Callable


def is_filter(filter_name: str) -> Callable:
    def outer_wrapper(func):
        func.is_filter = True
        func.filter_name = filter_name

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    return outer_wrapper


def export_functions(module_name: str, attr: str) -> list[Callable]:
    module = sys.modules[module_name]
    funcs = []

    for member in dir(module):
        export = getattr(module, member)
        applicable_export = getattr(export, attr, False)
        if applicable_export:
            funcs.append(export)

    return funcs



import logging
from uuid import UUID
from flask import (
    Blueprint, flash, g, redirect, render_template, request, session, url_for, make_response
)
from argus.backend.controller.notifications import bp as notifications_bp
from argus.backend.controller.team_ui import bp as teams_bp
from argus.backend.service.argus_service import ArgusService
from argus.backend.models.web import WebFileStorage
from argus.backend.service.user import UserService, login_required

LOGGER = logging.getLogger(__name__)

bp = Blueprint('main', __name__)
bp.register_blueprint(notifications_bp)
bp.register_blueprint(teams_bp)


@bp.route("/test_runs")
@login_required
def test_runs():
    return render_template("test_runs.html.j2")


@bp.route("/test_run/<string:run_id>")
@login_required
def test_run(run_id: UUID):
    return render_template("test_run.html.j2", id=run_id)


@bp.route("/test/<string:test_id>/runs")
@login_required
def runs(test_id: UUID):
    additional_runs = request.args.getlist("additionalRuns[]")
    return render_template("standalone_test_with_runs.html.j2", test_id=test_id, additional_runs=additional_runs)


@bp.route("/")
def home():
    return redirect(url_for("main.run_dashboard"))


@bp.route("/run_dashboard")
@bp.route("/workspace")
@login_required
def run_dashboard():
    return render_template('dashboard.html.j2')


@bp.route("/releases")
@login_required
def releases():
    service = ArgusService()
    all_releases = service.get_releases()
    return render_template("releases.html.j2", releases=all_releases)


@bp.route("/alert_debug")
@login_required
def alert_debug():
    alert_type = request.args.get("type", "success")
    message = request.args.get("message", "No message provided")
    flash(message=message, category=alert_type)
    return render_template("flash_debug.html.j2")


@bp.route("/dashboard/<string:release_name>")
@login_required
def release_dashboard(release_name: str):
    service = ArgusService()
    release, release_groups, release_tests = service.get_data_for_release_dashboard(
        release_name=release_name)
    data_json = {
        "release": dict(release.items()),
        "groups": [dict(group.items()) for group in release_groups],
        "tests": [dict(test.items()) for test in release_tests],
    }
    return render_template("release_dashboard.html.j2", release_name=release_name, data=data_json)


@bp.route("/release/<string:name>/scheduler")
@login_required
def release_scheduler(name: str):
    service = ArgusService()
    release, release_groups, release_tests = service.get_data_for_release_dashboard(
        release_name=name)
    data_json = {
        "release": dict(release.items()),
        "groups": [dict(group.items()) for group in release_groups],
        "tests": [dict(test.items()) for test in release_tests],
    }
    return render_template("release_schedule.html.j2", release_name=name, data=data_json)


@bp.route("/release/<string:name>/duty")
@login_required
def duty_planner(name: str):
    service = ArgusService()
    release, release_groups, release_tests = service.get_data_for_release_dashboard(
        release_name=name)
    data_json = {
        "release": dict(release.items()),
        "groups": [dict(group.items()) for group in release_groups],
        "tests": [dict(test.items()) for test in release_tests],
    }
    return render_template("duty_planner.html.j2", release_name=name, data=data_json)


@bp.route("/error/")
def error():
    return render_template("error.html.j2", type=request.args.get("type", 400))


@bp.route("/profile/")
@login_required
def profile():
    first_run = session.pop("first_run_info", None)
    token_generated = session.pop("token_generated", None)

    return render_template("profile.html.j2", first_run=first_run, token_generated=token_generated)


@bp.route("/profile/oauth/github", methods=["GET"])
def profile_oauth_github_callback():
    req_state = request.args.get('state', '')
    if req_state != session["csrf_token"]:
        return redirect(url_for("main.error", type=403))

    req_code = request.args.get("code", "WTF")
    service = UserService()
    try:
        first_run_info = service.github_callback(req_code)
    except Exception as exc:  # pylint: disable=broad-except
        flash(message=exc.args[0], category="error")
        return redirect(url_for("main.error", type=403))
    if first_run_info:
        session["first_run_info"] = first_run_info

    return redirect(url_for("main.profile"))


@bp.route("/storage/picture/<string:picture_id>")
@login_required
def get_picture(picture_id: str):
    res = make_response()
    try:
        picture = WebFileStorage.get(id=picture_id)
        with open(picture.filepath, "rb") as file:
            res.set_data(file.read())
        res.content_type = "image/*"
        res.status = 200
    except FileNotFoundError:  # pylint: disable=broad-except
        res.status = 404
        res.content_type = "text/plain"
        res.set_data("404 NOT FOUND")

    res.cache_control.max_age = 86400
    res.cache_control.public = True
    return res


@bp.route("/profile/update/picture", methods=["POST"])
@login_required
def upload_file():
    req_file = request.files.get("filedata")
    picture_data = req_file.stream.read()
    picture_name = req_file.filename
    if not req_file.content_type.startswith("image/"):
        flash(
            message=f"Expected image/*, got {req_file.content_type}", category="error")
        return redirect(url_for("main.profile"))
    if not picture_data:
        flash(message="No picture provided", category="error")
        return redirect(url_for("main.profile"))

    service = UserService()
    filename, filepath = service.save_profile_picture_to_disk(
        picture_name, picture_data, g.user.username)
    service.update_profile_picture(filename, filepath)

    return redirect(url_for("main.profile"))


@bp.route("/profile/update/name", methods=["POST"])
@login_required
def update_full_name():
    new_name = request.values.get("new_name")
    if not new_name:
        flash(message="Incorrect new name", category="error")
    else:
        service = UserService()
        service.update_name(g.user, new_name)
        flash("Successfully changed name!", category="success")
    return redirect(url_for("main.profile"))


@bp.route("/profile/update/email", methods=["POST"])
@login_required
def update_email():
    new_email = request.values.get("new_email")
    if not new_email:
        flash("Incorrect new email", category="error")
    else:
        service = UserService()
        service.update_email(g.user, new_email)
        flash("Successfully changed email!", category="success")
    return redirect(url_for("main.profile"))


@bp.route("/profile/update/password", methods=["POST"])
@login_required
def update_password():
    old_password = request.values.get("old_password")
    new_password = request.values.get("new_password")
    new_password_confirm = request.values.get("new_password_confirm")
    if not old_password:
        flash("Old password wasn't provided", category="error")
        return redirect(url_for("main.profile"))
    if not new_password:
        flash("New password wasn't provided", category="error")
        return redirect(url_for("main.profile"))

    if not new_password == new_password_confirm:
        flash("New password doesn't match confirmation!", category="error")
        return redirect(url_for("main.profile"))

    service = UserService()
    try:
        service.update_password(
            g.user, old_password=old_password, new_password=new_password)
    except Exception:  # pylint: disable=broad-except
        flash("Old password is incorrect", category="error")
        return redirect(url_for("main.profile"))

    flash("Successfully changed password!")
    return redirect(url_for("main.profile"))


@bp.route("/profile/jobs", methods=["GET"])
@login_required
def profile_jobs():
    service = ArgusService()
    jobs = service.get_jobs_for_user(g.user)
    return render_template("profile_jobs.html.j2", runs=jobs)


@bp.route("/profile/schedules", methods=["GET"])
@login_required
def profile_schedules():
    service = ArgusService()
    schedules = service.get_schedules_for_user(g.user)
    return render_template("profile_schedules.html.j2", schedules=schedules)
